% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconc_gaussian.R
\name{reconc_gaussian}
\alias{reconc_gaussian}
\title{Analytical reconciliation of Gaussian base forecasts}
\usage{
reconc_gaussian(S, base_forecasts.mu, base_forecasts.Sigma)
}
\arguments{
\item{S}{summing matrix (n x n_bottom).}

\item{base_forecasts.mu}{a vector containing the means of the base forecasts.}

\item{base_forecasts.Sigma}{a matrix containing the covariance matrix of the base forecasts.}
}
\value{
A list containing the bottom reconciled forecasts. The list has the following named elements:
\itemize{
\item \code{bottom_reconciled_mean}: reconciled mean for the bottom forecasts;
\item \code{bottom_reconciled_covariance}: reconciled covariance for the bottom forecasts.
}
}
\description{
Closed form computation of the reconciled forecasts in case of Gaussian base forecasts.
}
\details{
The order of the base forecast means and covariance is given by the order of the time series in the summing matrix.

The function returns only the reconciled parameters of the bottom variables.
The reconciled upper parameters and the reconciled samples for the entire hierarchy can be obtained from the reconciled bottom parameters.
See the example section.
}
\examples{

library(bayesRecon)

# Create a minimal hierarchy with 2 bottom and 1 upper variable
rec_mat <- get_reconc_matrices(agg_levels=c(1,2), h=2)
S <- rec_mat$S
A <- rec_mat$A

#Set the parameters of the Gaussian base forecast distributions
mu1 <- 2
mu2 <- 4
muY <- 9
mus <- c(muY,mu1,mu2)

sigma1 <- 2
sigma2 <- 2
sigmaY <- 3
sigmas <- c(sigmaY,sigma1,sigma2)

Sigma <- diag(sigmas^2)  #need to transform into covariance matrix
analytic_rec <- reconc_gaussian(S, base_forecasts.mu = mus,
                               base_forecasts.Sigma = Sigma)

bottom_mu_reconc <- analytic_rec$bottom_reconciled_mean
bottom_Sigma_reconc <- analytic_rec$bottom_reconciled_covariance

# Obtain reconciled mu and Sigma for the upper variable
upper_mu_reconc <- A \%*\% bottom_mu_reconc
upper_Sigma_reconc <- A \%*\% bottom_Sigma_reconc \%*\% t(A)

# Obtain reconciled mu and Sigma for the entire hierarchy
Y_mu_reconc <- S \%*\% bottom_mu_reconc
Y_Sigma_reconc <- S \%*\% bottom_Sigma_reconc \%*\% t(S)  # note: singular matrix

# Obtain reconciled samples for the entire hierarchy:
# i.e., sample from the reconciled bottoms and multiply by S
chol_decomp = chol(bottom_Sigma_reconc) # Compute the Cholesky Decomposition
Z = matrix(stats::rnorm(n = 2000), nrow = 2) # Sample from standard normal
B = t(chol_decomp) \%*\% Z + matrix(rep(bottom_mu_reconc, 1000), nrow=2) # Apply the transformation

U = S \%*\% B
Y_reconc = rbind(U, B)

}
\references{
Corani, G., Azzimonti, D., Augusto, J.P.S.C., Zaffalon, M. (2021).
\emph{Probabilistic Reconciliation of Hierarchical Forecast via Bayes' Rule}.
ECML PKDD 2020. Lecture Notes in Computer Science, vol 12459.
\doi{10.1007/978-3-030-67664-3_13}.

Zambon, L., Agosto, A., Giudici, P., Corani, G. (2024).
\emph{Properties of the reconciled distributions for Gaussian and count forecasts}.
International Journal of Forecasting (in press).
\doi{10.1016/j.ijforecast.2023.12.004}.
}
\seealso{
\code{\link[=reconc_BUIS]{reconc_BUIS()}}
}
